import math
from datetime import datetime
from unittest import mock

import boto3
from moto import mock_aws

from osc_ingest_trino import (
    TrinoBatchInsert,
    attach_trino_engine,
    pandas_type_to_sql,
    sql_compliant_name,
)
from osc_ingest_trino.unmanaged import unmanaged_parquet_tabledef

# from os_bucket import OSBucket


class MyModel:
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def save(self):
        s3 = boto3.client("s3", region_name="us-east-1")
        s3.put_object(Bucket="mybucket", Key=self.name, Body=self.value)


def test_sql_compliant_name():
    foo_name = sql_compliant_name("foo")
    assert foo_name == "foo"


def test_pandas_type_to_sql():
    integer_type = pandas_type_to_sql("int32")
    assert integer_type == "integer"


@mock.patch("osc_ingest_trino.trino_utils.trino.auth.JWTAuthentication")
@mock.patch("osc_ingest_trino.trino_utils.create_engine")
def test_attach_trino_engine(mock_engine, mock_trino_auth, monkeypatch):
    monkeypatch.setenv("TEST_USER", "tester")
    monkeypatch.setenv("TEST_PASSWD", "supersecret123")
    monkeypatch.setenv("TEST_HOST", "example")
    monkeypatch.setenv("TEST_PORT", "8000")

    fake_engine = mock.MagicMock()
    fake_engine.connect.return_value = None
    mock_engine.return_value = fake_engine
    mock_trino_auth.return_value = "yep"

    attach_trino_engine(
        env_var_prefix="TEST", catalog="ex_catalog", schema="ex_schema", verbose=True
    )

    mock_engine.assert_called_with(
        "trino://tester@example:8000/ex_catalog/ex_schema",
        connect_args={"auth": "yep", "http_scheme": "https"},
    )


# from os_bucket import OSBucket


def test_trino_batch_insert():
    # mock up an sqlalchemy table
    tbl = mock.MagicMock()
    tbl.name = "test"
    # mock up an sqlalchemy Connnection
    cxn = mock.MagicMock()
    # tuple data, in form supplied to __call__ as specified in 'method' param docs:
    # https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_sql.html
    rows = [
        ("a", 4.5),
        ("b'c", math.nan),
        (None, math.inf),
        ("d", -math.inf),
        ("e", datetime(2022, 1, 1)),
        (":f", 1.0),
    ]
    # invoke the __call__ method, simulating df.to_sql call
    tbi = TrinoBatchInsert(
        catalog="test", schema="test", batch_size=2, verbose=True, optimize=True
    )
    tbi(tbl, cxn, [], rows)

    assert cxn.execute.call_count == 4
    xcalls = cxn.execute.call_args_list
    assert (
        xcalls[0].args[0].text
        == "insert into test.test.test values\n('a', 4.5),\n('b''c', nan())"
    )
    assert (
        xcalls[1].args[0].text
        == "insert into test.test.test values\n(NULL, infinity()),\n('d', -infinity())"
    )
    assert (
        xcalls[2].args[0].text
        == "insert into test.test.test values\n('e', TIMESTAMP '2022-01-01 00:00:00'),\n('\\:f', 1.0)"
    )
    assert xcalls[3].args[0].text == "alter table test.test.test execute optimize"


def test_trino_pandas_insert():
    import pandas as pd

    # mock up an sqlalchemy table
    tbl = mock.MagicMock()
    tbl.name = "test_pandas"
    # mock up an sqlalchemy Connnection
    cxn = mock.MagicMock()
    df = pd.DataFrame(
        {
            "A": [4.5],
            "B'C": [math.nan],
            None: [math.inf],
            "D": [-math.inf],
            "E": [datetime(2022, 1, 1)],
            ":F": [1.0],
        }
    ).convert_dtypes()
    assert (
        df.dtypes
        == ["Float64", "Int64", "Float64", "Float64", "datetime64[ns]", "Int64"]
    ).all()
    # This passes Mock test, but fails when used in Trino/Iceberg environment
    df.to_sql(
        tbl.name,
        con=cxn,
        schema="test",
        if_exists="append",
        index=False,
        method=TrinoBatchInsert(batch_size=5, verbose=True),
    )


@mock_aws
def test_unmanaged_parquet_tabledef():
    import pandas as pd

    df = pd.DataFrame({"a": [1, 2, 3], "b": ["x", "y", "z"]}).convert_dtypes()

    conn = boto3.resource("s3", region_name="us-east-1")
    # We need to create the bucket since this is all in Moto's 'virtual' AWS account
    bucket = conn.Bucket("mybucket")
    bucket.create()

    tabledef = unmanaged_parquet_tabledef(
        df, "catalog", "schema", "table", bucket, partition_columns=["a", "b"]
    )
    print(tabledef)



"""Trino interoperability functions."""

import math
import os
import uuid
from datetime import datetime
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import pandas as pd
import sqlalchemy
import trino
from mypy_boto3_s3.service_resource import Bucket
from sqlalchemy import Connection, Engine, Row, Table, create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.sql import text

import osc_ingest_trino as osc
import osc_ingest_trino.unmanaged as oscu

# from sqlalchemy.sql.schema import Table as Table


__all__ = [
    "attach_trino_engine",
    "fast_pandas_ingest_via_hive",
    "TrinoBatchInsert",
    "_do_sql",
]


def attach_trino_engine(
    env_var_prefix: str = "TRINO",
    catalog: Optional[str] = None,
    schema: Optional[str] = None,
    verbose: Optional[bool] = False,
) -> Engine:
    """Return a SQLAlchemy engine object representing a Trino instance.

    env_var_prefix -- a prefix for all environment variables related to the Trino instance.
    catalog -- the Trino catalog.
    schema -- the Trino schema.
    verbose -- if True, print the full string used to connect.
    """
    sqlstring = "trino://{user}@{host}:{port}".format(
        user=os.environ[f"{env_var_prefix}_USER"],
        host=os.environ[f"{env_var_prefix}_HOST"],
        port=os.environ[f"{env_var_prefix}_PORT"],
    )
    if catalog is not None:
        sqlstring += f"/{catalog}"
    if schema is not None:
        if catalog is None:
            raise ValueError("connection schema specified without a catalog")
        sqlstring += f"/{schema}"

    sqlargs = {
        "auth": trino.auth.JWTAuthentication(os.environ[f"{env_var_prefix}_PASSWD"]),
        "http_scheme": "https",
    }

    if verbose:
        print(f"using connect string: {sqlstring}")

    engine = create_engine(sqlstring, connect_args=sqlargs)
    engine.connect()
    return engine


def _do_sql(
    sql: Union[sqlalchemy.sql.elements.TextClause, str],
    engine: Engine,
    verbose: bool = False,
) -> Optional[Sequence[Row[Any]]]:
    """Execute SQL query, returning the query result.

    sql -- the SQL query.
    engine -- the SQLAlchemy engine representing the Trino database.
    verbose -- if True, print the values returned from executing the string.
    """
    if type(sql) is not sqlalchemy.sql.elements.TextClause:
        sql = text(str(sql))
    if verbose:
        print(sql)
    with engine.begin() as cxn:
        qres = cxn.execute(sql)
    res = None
    if qres.returns_rows:
        res = qres.fetchall()
        if verbose:
            print(res)
    return res


def fast_pandas_ingest_via_hive(  # noqa: C901
    df: pd.DataFrame,
    engine: Engine,
    catalog: str,
    schema: str,
    table: str,
    hive_bucket: Bucket,
    hive_catalog: str,
    hive_schema: str,
    partition_columns: List[str] = [],
    overwrite: bool = False,
    typemap: Dict[str, str] = {},
    colmap: Dict[str, str] = {},
    verbose: bool = False,
) -> None:
    """Efficiently export a dataframe into a Trino database.

    df -- the dataframe to export.
    engine -- the SQLAlchemy engine representing the Trino database.
    catalog -- the Trino catalog.
    schema -- the Trino schema.
    table -- the name of the table created in the schema.
    hive_bucket -- the backing store of the Hive metastore.
    hive_catalog -- the Hive metastore catalog (where schemas are created).
    hive_schema -- the Hive metastore schema (where tables will be created).
    partition_columns -- if not empty, defines the partition columns of the table created.
    overwrite -- if True, an existing table will be overwritten.
    typemap -- used to format types that cannot otherwise be properly inferred.
    colmap -- used to format column names that cannot otherwise be properly inferred.
    verbose -- if True, print the queries being executed and the results of those queries.
    """
    uh8 = uuid.uuid4().hex[:8]
    hive_table = f"ingest_temp_{uh8}"

    dfw = df
    if len(partition_columns) > 0:
        if verbose:
            print("enforcing dataframe partition column order")
        dfw = osc.enforce_partition_column_order(dfw, partition_columns)

    # do this after any changes to DF column orderings above
    columnschema = osc.create_table_schema_pairs(dfw, typemap=typemap, colmap=colmap)

    # verify destination table first, to fail early and avoid creation of hive tables
    fq_tablename = ".".join([catalog, schema, table][(catalog is None) :])
    if verbose:
        print(f"\nverifying existence of table {fq_tablename}")
    tabledef = f"create table if not exists {fq_tablename} (\n"
    tabledef += f"{columnschema}\n"
    tabledef += ") with (\n    format = 'parquet'"
    if len(partition_columns) > 0:
        tabledef += ",\n"
        tabledef += f"    partitioning = array{partition_columns}"
    tabledef += "\n)"
    _do_sql(tabledef, engine, verbose=verbose)

    if verbose:
        print(f"\nstaging dataframe parquet to s3 {hive_bucket.name}")
    oscu.ingest_unmanaged_parquet(
        dfw,
        hive_schema,
        hive_table,
        hive_bucket,
        partition_columns=partition_columns,
        verbose=verbose,
    )

    try:
        if verbose:
            print(
                f"\ndeclaring intermediate hive table {hive_catalog}.{hive_schema}.{hive_table}"
            )
        tabledef = (
            f"create table if not exists {hive_catalog}.{hive_schema}.{hive_table} (\n"
        )
        tabledef += f"{columnschema}\n"
        tabledef += ") with (\n    format = 'parquet',\n"
        if len(partition_columns) > 0:
            tabledef += f"    partitioned_by = array{partition_columns},\n"
        tabledef += f"    external_location = 's3a://{hive_bucket.name}/trino/{hive_schema}/{hive_table}/'\n)"
        _do_sql(tabledef, engine, verbose=verbose)

        if verbose:
            print("\nsyncing partition metadata on intermediate hive table")
        if len(partition_columns) > 0:
            sql = text(
                f"call {hive_catalog}.system.sync_partition_metadata('{hive_schema}', '{hive_table}', 'FULL')"
            )
            _do_sql(sql, engine, verbose=verbose)

        if overwrite:
            if verbose:
                print(f"\noverwriting data in {fq_tablename}")
            sql = text(f"delete from {fq_tablename}")
            _do_sql(sql, engine, verbose=verbose)

        if verbose:
            print(
                f"\ntransferring data: {hive_catalog}.{hive_schema}.{hive_table} -> {fq_tablename}"
            )
        sql = text(
            f"insert into {fq_tablename}\nselect * from {hive_catalog}.{hive_schema}.{hive_table}"
        )
        _do_sql(sql, engine, verbose=verbose)

        if verbose:
            print(
                f"\ndeleting table and data for intermediate table {hive_catalog}.{hive_schema}.{hive_table}"
            )
        oscu.drop_unmanaged_table(
            hive_catalog, hive_schema, hive_table, engine, hive_bucket, verbose=verbose
        )
    except SQLAlchemyError:
        # Clean up table that will otherwise be orphaned
        if verbose:
            print(
                f"\ndeleting table and data for intermediate table {hive_catalog}.{hive_schema}.{hive_table}"
            )
        oscu.drop_unmanaged_table(
            hive_catalog, hive_schema, hive_table, engine, hive_bucket, verbose=verbose
        )
        raise


class TrinoBatchInsert(object):
    """A class used to bundle together basic Trino parameters."""

    def __init__(
        self,
        catalog: Optional[str] = None,
        schema: Optional[str] = None,
        batch_size: int = 1000,
        optimize: bool = False,
        verbose: bool = False,
    ):
        """Initialize TrinoBatchInsert objects."""
        self.catalog = catalog
        self.schema = schema
        self.batch_size = batch_size
        self.optimize = optimize
        self.verbose = verbose

    # conforms to signature expected by pandas 'callable' value for method kw arg
    # https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_sql.html
    # https://pandas.pydata.org/docs/user_guide/io.html#io-sql-method
    def __call__(
        self,
        sqltbl: Table,
        dbcxn: Connection,
        columns: List[str],
        data_iter: List[Tuple],
    ) -> None:
        """Implement `callable` interface for row-by-row insertion."""
        fqname = self._full_table_name(sqltbl)
        batch: List[str] = []
        self.ninserts = 0
        for r in data_iter:
            # each row of data_iter is a python tuple
            # I cannot currently make good use of sqlalchemy ':params'
            # and so I have to do my own "manual" value formatting for insertions
            row = ", ".join([TrinoBatchInsert._sqlform(e) for e in r])
            row = f"({row})"
            batch.append(row)
            # possible alternative: dispatch batches by total batch size in bytes
            if len(batch) >= self.batch_size:
                self._do_insert(dbcxn, fqname, batch)
                batch = []
        if len(batch) > 0:
            self._do_insert(dbcxn, fqname, batch)
        if self.optimize:
            if self.verbose:
                print("optimizing table files")
            sql = text(f"alter table {fqname} execute optimize")
            qres = dbcxn.execute(sql)
            x = qres.fetchall()
            if self.verbose:
                print(f"execute optimize: {x}")

    def _do_insert(self, dbcxn: Connection, fqname: str, batch: List[str]) -> None:
        """Implement actual row-by-row insertion of BATCH data into table FQNAME using DBCXN database connection."""
        if self.verbose:
            print(f"inserting {len(batch)} records")
            TrinoBatchInsert._print_batch(batch)
        # trino is not currently supporting sqlalchemy cursor.executemany()
        # and so I am generating an insert command with no ':params' that
        # includes all batch data as literal sql values
        valclause = ",\n".join(batch)
        # injecting raw sql strings is deprecated and will be illegal in sqlalchemy 2.x
        # using text() is the correct way:
        sql = text(f"insert into {fqname} values\n{valclause}")
        # if self.verbose: print(f'{sql}')
        qres = dbcxn.execute(sql)
        x = qres.fetchall()
        if self.verbose:
            print(f"batch insert result: {x}")

    def _full_table_name(self, sqltbl: Table) -> str:
        """Return fully qualified table name for SQLTBL table within this TrinoBatchInsert object."""
        # start with table name
        name: str = f"{sqltbl.name}"
        # prepend schema - allow override from this class
        name = f"{self.schema or sqltbl.schema}.{name}"
        # prepend catalog, if provided
        if self.catalog is not None:
            name = f"{self.catalog}.{name}"
        if self.verbose:
            print(f'constructed fully qualified table name as: "{name}"')
        return name

    @staticmethod
    def _sqlform(x: Any) -> str:
        """Format the value of x so it can appear in a SQL Values context."""
        if x is None:
            return "NULL"
        if isinstance(x, str):
            # escape any single quotes in the string
            t = x.replace("'", "''")
            # colons are mostly a problem for ':some_id_name', which is interpreted as
            # a parameter requiring binding, but just escaping them all works
            t = t.replace(":", "\\:")
            # enclose string with single quotes
            return f"'{t}'"
        if isinstance(x, datetime):
            return f"TIMESTAMP '{x}'"
        if isinstance(x, float):
            if math.isnan(x):
                return "nan()"
            if math.isinf(x):
                if x < 0:
                    return "-infinity()"
                return "infinity()"
        return str(x)

    @staticmethod
    def _print_batch(batch: List[str]) -> None:
        """For batch, a list of SQL query lines, print up to the first 5 such."""
        if len(batch) > 5:
            print("\n".join(f"  {e}" for e in batch[:3]))
            print("  ...")
            print(f"  {batch[-1]}")
        else:
            print("\n".join(f"  {e}" for e in batch))



"""Functions to create, ingest, and drop unmanaged Hive tables."""

import shutil
import uuid

import pandas as pd
from sqlalchemy import text

from osc_ingest_trino import create_table_schema_pairs, upload_directory_to_s3

__all__ = [
    "drop_unmanaged_table",
    "drop_unmanaged_data",
    "ingest_unmanaged_parquet",
    "unmanaged_parquet_tabledef",
]

_default_prefix = "trino/{schema}/{table}"


def _remove_trailing_slash(s):
    """Remove trailing slash from s."""
    s = str(s)
    if len(s) == 0:
        return s
    if s[-1] != "/":
        return s
    return _remove_trailing_slash(s[:-1])


def _prefix(pfx, schema, table):
    """Translate pfx, schema, and table names into S3 bucket name."""
    return _remove_trailing_slash(pfx).format(schema=schema, table=table)


def drop_unmanaged_table(
    catalog, schema, table, engine, bucket, prefix=_default_prefix, verbose=False
):
    """Drop catalog.schema.table from Hive metastore and also delete its S3 backing store."""
    sql = text(f"drop table if exists {catalog}.{schema}.{table}")
    with engine.begin() as cxn:
        qres = cxn.execute(sql)
    dres = bucket.objects.filter(Prefix=f"{_prefix(prefix, schema, table)}/").delete()
    if verbose:
        print(dres)
    return qres


def drop_unmanaged_data(schema, table, bucket, prefix=_default_prefix, verbose=False):
    """Delete data that may have been orphaned when its table was dropped in Hive metastore."""
    dres = bucket.objects.filter(Prefix=f"{_prefix(prefix, schema, table)}/").delete()
    if verbose:
        print(dres)
    return dres


def ingest_unmanaged_parquet(
    df,
    schema,
    table,
    bucket,
    partition_columns=[],
    append=True,
    workdir="/tmp",
    prefix=_default_prefix,
    verbose=False,
):
    """Ingest data from df into Hive metastore table with backing store bucket."""
    if not isinstance(df, pd.DataFrame):
        raise ValueError("df must be a pandas DataFrame")
    if not isinstance(partition_columns, list):
        raise ValueError("partition_columns must be list of column names")

    s3pfx = _prefix(prefix, schema, table)

    if not append:
        dres = bucket.objects.filter(Prefix=f"{s3pfx}/").delete()
        if verbose:
            print(dres)

    if len(partition_columns) > 0:
        # tell pandas to write a directory tree, using partitions
        tmp = f"{workdir}/{table}"
        # pandas does not clean out destination directory for you:
        shutil.rmtree(tmp, ignore_errors=True)
        df.to_parquet(tmp, partition_cols=partition_columns, index=False)
        # upload the tree onto S3
        upload_directory_to_s3(tmp, bucket, s3pfx, verbose=verbose)
    else:
        # do not use partitions: a single parquet file is created
        parquet = f"{uuid.uuid4().hex}.parquet"
        tmp = f"{workdir}/{parquet}"
        df.to_parquet(tmp, index=False)
        dst = f"{s3pfx}/{parquet}"
        if verbose:
            print(f"{tmp}  -->  {dst}")
        bucket.upload_file(tmp, dst)


def unmanaged_parquet_tabledef(
    df,
    catalog,
    schema,
    table,
    bucket,
    partition_columns=[],
    typemap={},
    colmap={},
    verbose=False,
):
    """Return a SQL string that would create a table suitable for ingesting df into Hive metastore backed by bucket."""
    if not isinstance(df, pd.DataFrame):
        raise ValueError("df must be a pandas DataFrame")
    if not isinstance(partition_columns, list):
        raise ValueError("partition_columns must be list of column names")

    columnschema = create_table_schema_pairs(df, typemap=typemap, colmap=colmap)

    tabledef = f"create table if not exists {catalog}.{schema}.{table} (\n"
    tabledef += f"{columnschema}\n"
    tabledef += ") with (\n    format = 'parquet',\n"
    if len(partition_columns) > 0:
        tabledef += f"    partitioned_by = array{partition_columns},\n"
    tabledef += (
        f"    external_location = 's3a://{bucket.name}/trino/{schema}/{table}/'\n)"
    )

    if verbose:
        print(tabledef)
    return tabledef



"""Functions to translate Pandas column names to SQL column names."""

import re
from typing import List, Union, cast

import pandas as pd

__all__ = [
    "sql_compliant_name",
    "enforce_sql_column_names",
    "enforce_partition_column_order",
]

_wsdedup = re.compile(r"\s+")
_usdedup = re.compile(r"__+")
_rmpunc = re.compile(r"[,.()&$/+-]+")


# 63 seems to be a common max column name length
def sql_compliant_name(name: Union[List[str], str], maxlen=63) -> Union[List[str], str]:
    """Convert name to a SQL-compliant table or column name, abbreviating some common words."""
    if isinstance(name, list):
        return [cast(str, sql_compliant_name(e, maxlen=maxlen)) for e in name]
    w = str(name).casefold().rstrip().lstrip()
    w = w.replace("-", "_")
    w = _rmpunc.sub("", w)
    w = _wsdedup.sub("_", w)
    w = _usdedup.sub("_", w)
    w = w.replace("average", "avg")
    w = w.replace("maximum", "max")
    w = w.replace("minimum", "min")
    w = w.replace("absolute", "abs")
    w = w.replace("source", "src")
    w = w.replace("distribution", "dist")
    # these are common in the sample names but unsure of standard abbv
    # w = w.replace("inference", "inf")
    # w = w.replace("emissions", "emis")
    # w = w.replace("intensity", "int")
    # w = w.replace("reported", "rep")
    # w = w.replace("revenue", "rev")
    w = w[:maxlen]
    return w


def enforce_sql_column_names(
    df: pd.DataFrame, inplace: bool = False, maxlen: int = 63
) -> pd.DataFrame:
    """Ensure that all column names for df are SQL-compliant."""
    if not isinstance(df, pd.DataFrame):
        raise ValueError("df must be a pandas DataFrame")
    icols = df.columns.to_list()
    ocols = sql_compliant_name(icols, maxlen=maxlen)
    if len(set(ocols)) < len(ocols):
        raise ValueError("remapped column names were not unique!")
    rename_map = dict(list(zip(icols, ocols)))
    return df.rename(columns=rename_map, inplace=inplace)


def enforce_partition_column_order(
    df: pd.DataFrame, pcols: List[str], inplace: bool = False
) -> pd.DataFrame:
    """Reorder columns names of df to match the order given by pcols."""
    if not isinstance(df, pd.DataFrame):
        raise ValueError("df must be a pandas DataFrame")
    if not isinstance(pcols, list):
        raise ValueError("pcols must be list of column names")
    pcols = [str(e) for e in pcols]
    cols = list(df.columns.values)
    for c in pcols:
        cols.remove(c)
        cols.append(c)
    if not inplace:
        return df[cols]
    for c in cols:
        s = df[c]
        df.drop(columns=[c], inplace=True)
        df[c] = s
    return df



"""Functions to simplify use of S3-based data, Pandas dataframes, and Trino SQL tables."""

from .boto3_utils import attach_s3_bucket, upload_directory_to_s3
from .dotenv_utils import load_credentials_dotenv
from .sqlcols import (
    enforce_partition_column_order,
    enforce_sql_column_names,
    sql_compliant_name,
)
from .sqltypes import create_table_schema_pairs, pandas_type_to_sql
from .trino_utils import (
    TrinoBatchInsert,
    _do_sql,
    attach_trino_engine,
    fast_pandas_ingest_via_hive,
)

__all__ = [
    "sql_compliant_name",
    "enforce_sql_column_names",
    "enforce_partition_column_order",
    "pandas_type_to_sql",
    "create_table_schema_pairs",
    "attach_s3_bucket",
    "upload_directory_to_s3",
    "load_credentials_dotenv",
    "attach_trino_engine",
    "fast_pandas_ingest_via_hive",
    "TrinoBatchInsert",
    "_do_sql",
]



"""AWS S3 interoperability functions."""

import os
from pathlib import Path
from typing import Union

import boto3
from mypy_boto3_s3.service_resource import Bucket

__all__ = [
    "upload_directory_to_s3",
    "attach_s3_bucket",
]


def upload_directory_to_s3(
    path: Union[Path, str], bucket: Bucket, prefix: str, verbose: bool = False
) -> None:
    """Upload files to an S3 bucket.

    path -- the directory containing all files to be uploaded.
    bucket -- the S3 bucket.
    prefix -- the prefix prepended to each filename before uploading.
    verbose -- if True, print each file uploaded (with its prefix).
    """
    path = str(path)
    prefix = str(prefix)
    for subdir, dirs, files in os.walk(path):
        for f in files:
            pfx = subdir.replace(path, prefix)
            src = os.path.join(subdir, f)
            dst = os.path.join(pfx, f)
            if verbose:
                print(f"{src}  -->  {dst}")
            bucket.upload_file(src, dst)


def attach_s3_bucket(env_var_prefix: str) -> Bucket:
    """Return the S3 Bucket resource asscoiated with env_var_prefix (typically from `credentials.env`)."""
    s3 = boto3.resource(
        service_name="s3",
        endpoint_url=os.environ[f"{env_var_prefix}_ENDPOINT"],
        aws_access_key_id=os.environ[f"{env_var_prefix}_ACCESS_KEY"],
        aws_secret_access_key=os.environ[f"{env_var_prefix}_SECRET_KEY"],
    )
    return s3.Bucket(os.environ[f"{env_var_prefix}_BUCKET"])



"""Functions to translate Pandas dataframes to SQL equivalents."""

from typing import Dict

import pandas as pd

__all__ = [
    "pandas_type_to_sql",
    "create_table_schema_pairs",
]

_p2smap = {
    "string": "varchar",
    "float32": "real",
    "Float32": "real",
    "float64": "double",
    "Float64": "double",
    "int32": "integer",
    "Int32": "integer",
    "int64": "bigint",
    "Int64": "bigint",
    "bool": "boolean",
    "boolean": "boolean",
    "category": "varchar",
    "datetime64[ns, UTC]": "timestamp",
}


def pandas_type_to_sql(pt: str, typemap: Dict[str, str] = {}):
    """Return the SQL type corresponding to the pandas type `pt` (using special mappings, if any, from `typemap`)."""
    if not isinstance(typemap, dict):
        raise ValueError("typemap must be a dict")
    # user defined typemap overrides _p2smap
    st = typemap.get(pt, _p2smap.get(pt))
    if st is not None:
        return st
    raise ValueError("unexpected pandas column type '{pt}'".format(pt=pt))


def create_table_schema_pairs(
    df: pd.DataFrame,
    typemap: Dict[str, str] = {},
    colmap: Dict[str, str] = {},
    indent: int = 4,
) -> str:
    """Create SQL column, type pairs that can appear in a CREATE TABLE operation.

    df -- the dataframe to be rendered as a SQL table
    typemap -- mappings from dtypes to SQL types above and beyond our defaults
    colmap -- mappings of df column names to SQL column names if not using defaults
    indent -- how many spaces of indent to make our SQL declarations pretty
    """
    if not isinstance(df, pd.DataFrame):
        raise ValueError("df must be a pandas DataFrame")
    if not isinstance(colmap, dict):
        raise ValueError("colmap must be a dict")
    columns = df.columns.to_list()
    try:
        types = [
            colmap.get(col, pandas_type_to_sql(str(df[col].dtype), typemap=typemap))
            for col in columns
        ]
    except ValueError as exc:
        raise ValueError(f"df.dtypes\n{df.dtypes}\nraised {exc}")
    return ",\n".join([f"{' ' * indent}{e[0]} {e[1]}" for e in zip(columns, types)])



"""Functions create and clean up unmanaged Hive tables."""

from .unmanaged_hive_ingest import (
    drop_unmanaged_data,
    drop_unmanaged_table,
    ingest_unmanaged_parquet,
    unmanaged_parquet_tabledef,
)

__all__ = [
    "drop_unmanaged_table",
    "drop_unmanaged_data",
    "ingest_unmanaged_parquet",
    "unmanaged_parquet_tabledef",
]



"""Functions to read credentials files and inject secrets into the environment."""

import os
import pathlib

from dotenv import load_dotenv

__all__ = [
    "load_credentials_dotenv",
]


def load_credentials_dotenv() -> None:
    """Load some standard environment variables from a dot-env file, if it exists.

    If no such file can be found, do not raise, allowing these environment vars to be populated in some other way.
    """
    dotenv_dir = os.environ.get(
        "CREDENTIAL_DOTENV_DIR", os.environ.get("PWD", "/opt/app-root/src")
    )
    dotenv_path = pathlib.Path(dotenv_dir) / "credentials.env"
    if os.path.exists(dotenv_path):
        load_dotenv(dotenv_path=dotenv_path, override=True)
